/********************************************************************************
*                                                                               *
*                  U n d o / R e d o - a b l e   C o m m a n d                  *
*                                                                               *
*********************************************************************************
* Copyright (C) 2000 by Jeroen van der Zijp.   All Rights Reserved.             *
*********************************************************************************
* This library is free software; you can redistribute it and/or                 *
* modify it under the terms of the GNU Library General Public                   *
* License as published by the Free Software Foundation; either                  *
* version 2 of the License, or (at your option) any later version.              *
*                                                                               *
* This library is distributed in the hope that it will be useful,               *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU             *
* Library General Public License for more details.                              *
*                                                                               *
* You should have received a copy of the GNU Library General Public             *
* License along with this library; if not, write to the Free                    *
* Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.            *
*********************************************************************************
* $Id: FXUndoList.cpp,v 1.9 2000/07/18 19:43:39 jeroen Exp $                    *
********************************************************************************/
#include "xincs.h"
#include "fxver.h"
#include "fxdefs.h"
#include "FXStream.h"
#include "FXString.h"
#include "FXObject.h"
#include "FXObject.h"
#include "FXCharset.h"
#include "FXDict.h"
#include "FXRegistry.h"
#include "FXAccelTable.h"
#include "FXApp.h"
#include "FXId.h"
#include "FXDrawable.h"
#include "FXWindow.h"
#include "FXUndoList.h"

/*
  Notes:
  - Manages a simple undo/redo list.
  - Whenever adding a new command, the redo list is deleted.
  - When a command is undone, its moved to the redo list.
  - When a command is redone, its moved to the undo list.
  - At any time, you can trim down the undo list to a given
    maximum size (total undo information stored in all the records)
    or length (number of undo records).
  - You can call cut() to kill the redo list without adding a new command 
    to the undo list.
  - Need better mechanism for undo name or redo name.
  - Need to keep track of when to revert to "unmodified" state intelligently
    (e.g. saving a file after some number of commands, this means an undo
    makes the file "modified" again w.r.t. the saved copy on disk).
*/



/*******************************************************************************/

// Map
FXDEFMAP(FXUndoList) FXUndoListMap[]={
  FXMAPFUNC(SEL_COMMAND, FXUndoList::ID_CLEAR,      FXUndoList::onCmdClear),
  FXMAPFUNC(SEL_UPDATE,  FXUndoList::ID_CLEAR,      FXUndoList::onUpdClear),
  FXMAPFUNC(SEL_COMMAND, FXUndoList::ID_UNDO,       FXUndoList::onCmdUndo),
  FXMAPFUNC(SEL_UPDATE,  FXUndoList::ID_UNDO,       FXUndoList::onUpdUndo),
  FXMAPFUNC(SEL_COMMAND, FXUndoList::ID_REDO,       FXUndoList::onCmdRedo),
  FXMAPFUNC(SEL_UPDATE,  FXUndoList::ID_REDO,       FXUndoList::onUpdRedo),
  };


// Object implementation
FXIMPLEMENT(FXUndoList,FXObject,FXUndoListMap,ARRAYNUMBER(FXUndoListMap))

  
// Make new empty undo list
FXUndoList::FXUndoList(){
  redolist=NULL;
  undolist=NULL;
  size=0;
  count=0;
  }


// Cut the redo list
void FXUndoList::cut(){
  register FXCommand *p;
  while(redolist){
    p=redolist;
    redolist=redolist->next;
    delete p;
    }
  redolist=NULL;
  }


// Add new command, executing if desired
FXbool FXUndoList::add(FXCommand* command,FXbool doit){

  // Cut redo list
  cut();

  // Add to undo list
  command->next=undolist;
  undolist=command;

  // Execute it right now
  if(doit) command->redo();

  // Update info
  size+=command->size();
  count++;
  
  return TRUE;
  }


// Undo last command
FXbool FXUndoList::undo(){
  register FXCommand *command=undolist;
  if(command){
    command->undo();
    undolist=undolist->next;
    command->next=redolist;
    redolist=command;
    size-=command->size();
    count--;
    return TRUE;
    }
  return FALSE;
  }


// Redo next command
FXbool FXUndoList::redo(){
  register FXCommand *command=redolist;
  if(command){
    command->redo();
    redolist=redolist->next;
    command->next=undolist;
    undolist=command;
    size+=command->size();
    count++;
    return TRUE;
    }
  return TRUE;
  }


// Clear list
void FXUndoList::clear(){
  register FXCommand *p;
  while(redolist){
    p=redolist;
    redolist=redolist->next;
    delete p;
    }
  while(undolist){
    p=undolist;
    undolist=undolist->next;
    delete p;
    }
  redolist=NULL;
  undolist=NULL;
  count=0;
  size=0;
  }


// Clear undo list
long FXUndoList::onCmdClear(FXObject*,FXSelector,void*){
  clear();
  return 1;
  }


// Update Clear undo list
long FXUndoList::onUpdClear(FXObject* sender,FXSelector,void*){
  if(canUndo() || canRedo())
    sender->handle(this,MKUINT(FXWindow::ID_ENABLE,SEL_COMMAND),NULL);
  else
    sender->handle(this,MKUINT(FXWindow::ID_DISABLE,SEL_COMMAND),NULL);
  return 1;
  }


// Undo last command
long FXUndoList::onCmdUndo(FXObject*,FXSelector,void*){
  undo();
  return 1;
  }


// Update undo last command
long FXUndoList::onUpdUndo(FXObject* sender,FXSelector,void*){
  if(canUndo()){
    FXString string=undolist->undoName();
    sender->handle(this,MKUINT(FXWindow::ID_ENABLE,SEL_COMMAND),NULL);
    sender->handle(this,MKUINT(FXWindow::ID_SETSTRINGVALUE,SEL_COMMAND),(void*)&string);
    }
  else{
    sender->handle(this,MKUINT(FXWindow::ID_DISABLE,SEL_COMMAND),NULL);
    }
  return 1;
  }


// Redo last command
long FXUndoList::onCmdRedo(FXObject*,FXSelector,void*){
  redo();
  return 1;
  }


// Update redo last command
long FXUndoList::onUpdRedo(FXObject* sender,FXSelector,void*){
  if(canRedo()){
    FXString string=redolist->redoName();
    sender->handle(this,MKUINT(FXWindow::ID_ENABLE,SEL_COMMAND),NULL);
    sender->handle(this,MKUINT(FXWindow::ID_SETSTRINGVALUE,SEL_COMMAND),(void*)&string);
    }
  else{
    sender->handle(this,MKUINT(FXWindow::ID_DISABLE,SEL_COMMAND),NULL);
    }
  return 1;
  }



// Trim undo list down to at most n records
void FXUndoList::trimCount(FXuint nc){
  if(count>nc){
    register FXCommand **pp=&undolist;
    register FXCommand *p=*pp;
    register FXuint i=0;
    while(p && i<nc){ 
      pp=&p->next;
      p=*pp;
      i++; 
      }
    while(*pp){
      p=*pp;
      *pp=p->next;
      size-=p->size();
      count--;
      delete p;
      }
    }
  }


// Trim undo list down to at most size sz
void FXUndoList::trimSize(FXuint sz){
  if(size>sz){
    register FXCommand **pp=&undolist;
    register FXCommand *p=*pp;
    register FXuint s=0;
    while(p && (s=s+p->size())<=sz){ 
      pp=&p->next;
      p=*pp;
      }
    while(*pp){
      p=*pp;
      *pp=p->next;
      size-=p->size();
      count--;
      delete p;
      }
    }
  }


// Clean up
FXUndoList::~FXUndoList(){
  clear();
  }

